/**
 Author:  Daniel Archambault
 Email :  archam@cs.ubc.ca
 Last modification : 01/05/2006
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/

/**
 * implementation of TopoLayout.  For nodes, history, author and licence
 * please see the header file TopoLayout.h
 */

#include <cmath>
#include <algorithm>
#include <tulip/MetaGraphProxy.h>
#include <tulip/TlpTools.h>
#include <tulip/DrawingTools.h>
#include <tulip/SortIterator.h>
#include <tulip/StableIterator.h>
#include <tulip/TreeTest.h>
#include "GeoOperators.h"
#include "TopoLayout.h"
using namespace std;

LAYOUTPLUGIN(TopoLayout,"TopoLayout","Daniel Archambault","04/06/2006","Alpha","0","1");

//determines if a tree is bushy (degree > depth)
bool bushy (node n, node cameFrom, SuperGraph *featureGraph,
	    unsigned int curDepth, 
	    unsigned int &maxDepth, unsigned int &maxDegree,
	    MutableContainer<bool> &visited) {
  if (visited.get (n.id)) return maxDepth < maxDegree;
  visited.set (n.id, true);

  //see if we have a new max depth or max degree
  if (featureGraph->deg (n) > maxDegree) maxDegree = featureGraph->deg (n);
  if (curDepth > maxDepth) maxDepth = curDepth;
  
  //compute the max degree and always compare it to the depth
  Iterator <node> *it = featureGraph->getInOutNodes(n);
  while(it->hasNext()) {
    node curNode = it->next();
    if (curNode != cameFrom) 
      bushy (curNode, n, featureGraph, curDepth + 1, maxDepth, maxDegree,
	     visited);
  } delete it; //end while
  return maxDepth < maxDegree;
}//end bushy

//A test to determine if a tree is bushy or not
bool isBushy (SuperGraph *tree, TopoLayout *topoLayout) {
  node root;
  tlp::getSource (tree, root);
  topoLayout->setSpokeOrderAroundRoot (root);
  unsigned int maxDegree = 0, maxDepth = 0;
  MutableContainer<bool> visited;
  visited.setAll (false);
  return bushy (root, root, tree, 0, maxDepth, maxDegree, visited);
}//end isBushy

//A test to see if we have a high node/edge ratio
bool hasHighNodeEdgeRatio (SuperGraph *graph, TopoLayout *topoLayout) {
  return highNodeEdgeRatio (graph->numberOfNodes(), 
			    graph->numberOfEdges(),
			    topoLayout->nodeEdgeThresh);
}//end hasHighNodeEdgeRatio

//a function to find a graph with the passed name.  It returns
//NULL if the graph is not found
SuperGraph *TopoLayout::findGraph (string graphName) {
  SuperGraph *foundGraph = NULL;
  Iterator <SuperGraph *> *graphs = superGraph->getSubGraphs ();
  while (graphs->hasNext()) {
    SuperGraph *curGraph = graphs->next ();
    string name = curGraph->getAttribute <string> ("name");
    if (name == graphName) {
      foundGraph = curGraph;
      break;
    }//end if
  } delete graphs; //end while
  return foundGraph;
}//end findGraph

//a function to generate the originalEdges table at this level of the
//hierarchy
void TopoLayout::genOriginalEdges () {
  //a proxy for the original edge attribute
  StringProxy *originalEdgesProxy = 
    featureGraph->getProperty <StringProxy> (ORIGINAL_EDGE_NAME);
  
  //for all the edges at this level
  Iterator <edge> *edges = featureGraph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();
    
    //get the list of original edges and parse it.
    vector <edge> thisEdgeOrigEdge;

    //if there is a null string, there are no original edges
    //(we do not want to apply torques).
    if (originalEdgesProxy->getEdgeValue (curEdge) != "") {
      istringstream idparse (originalEdgesProxy->getEdgeValue (curEdge));
      unsigned int edgeid;
      while (!idparse.eof()) {
	idparse >> edgeid;
	thisEdgeOrigEdge.push_back (edge (edgeid));
      }//end while
    }//end if
    originalEdges.set (curEdge.id, thisEdgeOrigEdge);
  } delete edges; 
}//end genOriginalEdges

//A function to recursively lay out a graph hierarchy created by
//TopoClustering.  Returns true if it succeeds without error.
bool TopoLayout::recurseOnSubHierarchy (SuperGraph *curFeatureGraph,
					unsigned int &depth) {
  //for all the nodes of this hierarchy, recursively layout the
  //metagraph, if there is one
  Iterator <node> *hierarchyNodes = curFeatureGraph->getNodes();
  while (hierarchyNodes->hasNext()) {
    node curNode = hierarchyNodes->next();
    SuperGraph *feature = hierarchyMetaGraphs->getNodeValue (curNode);
    if (feature != NULL) {
      //layout all the metagraphs inside nodes of this graph
      unsigned int featureDepth = 0;
      bool success = recurseOnSubHierarchy (feature, featureDepth);
      if (!success) {
	string name = feature->getAttribute <string> ("name");
	cerr << "There was an error recursive call on hierarchy subgraph "
	     << name << endl;
	return false;
      }//end if
      if (depth < featureDepth) depth = featureDepth;

      //set the node size of this element of the hierarchy
      setNodeSize (curNode, curFeatureGraph);
    }//end if
  } delete hierarchyNodes; //end while
  ++depth;
  //layout the graph at this level
  bool success = layoutFeatureNode (curFeatureGraph);
  if (!success) {
    string name = hierarchy->getAttribute <string> ("name");
    cerr << "There was an error laying out the final hierarchy "
	 << name << endl;
    return false;
  }//end if
  return true;
}//end recurseOnSubHierarchy

//A function to draw the hierarchy
bool TopoLayout::drawHierarchy () {
  string errMsg; //for error messages
  
  //get the subgraphs and extract the hierarchy
  hierarchy = findGraph (HIERARCHY_SUBGRAPH_NAME);

  //if we cannot find the hierarchy, remake it
  if (hierarchy == NULL) {
    cerr << "Could not locate hierarchy!\n"
	 << "Launching TopoClustering...." << endl;
    
    //a local dataset for the clustering call
    DataSet topoClusterData;
    topoClusterData.set ("nodeEdgeThresh", nodeEdgeThresh);
    
    //call the topo cluster algorithm (passed DataSet should have hierarchy,
    //untouched, and finalLayoutGraph all undefined)
    bool success = 
      tlp::clusterizeGraph (superGraph, errMsg, 
			    &topoClusterData, 
			    "Topological Feature Clustering");
    if (!success) {
      cerr << "Topological Feature Clustering returned the following error:\n"
	   << errMsg << endl;
      cerr << "(is the plugin properly installed?)" << endl;
      return false;
    }//end if
    
    //get the generated graphs
    topoClusterData.get <SuperGraph *> ("hierarchy", hierarchy);
    topoClusterData.get <SuperGraph *> ("untouched", untouched);
    topoClusterData.get <SuperGraph *> ("finalLayoutGraph", finalLayoutGraph); 
  }//end if
  
  //otherwise the real hierarchy is one subgraph down
  else {
    Iterator <SuperGraph *> *hierSubGraph = hierarchy->getSubGraphs();
    hierarchy = hierSubGraph->next();
    delete hierSubGraph;
    
    //get final layout graph and initialize all of its coords to zero
    finalLayoutGraph = findGraph (FINAL_LAYOUT_SUBGRAPH_NAME);
    LayoutProxy *finalLayout = 
      finalLayoutGraph->getLocalProperty <LayoutProxy> ("viewLayout");
    finalLayout->setAllNodeValue (Coord (0, 0, 0));

    //get the untouched graph, and, don't touch it!
    untouched = findGraph (UNTOUCHED_SUBGRAPH_NAME);

    //set the final layout node sizes to the same sizes as untouched
    SizesProxy *finalSizes = 
      finalLayoutGraph->getLocalProperty <SizesProxy> ("viewSize");
    SizesProxy *untouchedSizes = 
      untouched->getProperty <SizesProxy> ("viewSize");
    Iterator <node> *nodes = superGraph->getNodes();
    while (nodes->hasNext()) {
      node curNode = nodes->next();
      finalSizes->setNodeValue (curNode, 
			      untouchedSizes->getNodeValue (curNode));
    } delete nodes; //end while
  }//end else
  hierarchyDepth = 0;
  return recurseOnSubHierarchy (hierarchy, hierarchyDepth);
}//end drawHierarchy

//A function to find the position of an original node of the graph
//at this level of the hierarchy.  That positition is returned 
//by the function.  The value of foundIn is set to the node in which
//originalNode was found.
Coord TopoLayout::findHierarchyPosition (node originalNode, 
					 node &foundIn) {
  //trace the source node up until we hit either the root
  //or the leaf
  Coord positionOfCur (0, 0, 0);
  LayoutProxy *atLevelLayout;
      
  //the current node and the graph that contains it
  node curNode = originalNode;
  SuperGraph *containedIn = containedInGraph->getNodeValue (curNode);

  //get the rotation proxy on the containedIn graph
  MetricProxy *rotations = 
    containedIn->getLocalProperty<MetricProxy> ("viewRotation");

  //while we have not reached the current level in the hierarchy
  while (containedIn != featureGraph) {
    //get a layout proxy for the graph at this level
    atLevelLayout = 
      containedIn->getLocalProperty<LayoutProxy> ("viewLayout");

    //apply the local translation and rotation to the position of cur
    positionOfCur += atLevelLayout->getNodeValue (curNode);
    positionOfCur = 
      rotate (rotations->getNodeValue (curNode), positionOfCur);
    
    //get the node that represents this graph
    unsigned int curNodeID = 
      containedIn->getAttribute <unsigned int> (PARENT_NODE_ATTRIBUTE);

    //if the curNodeID is NO_PARENT, we are seeking the contained
    //graph of a node that does not exist.  There is an error
    if (curNodeID == NO_PARENT) {
      cerr << "Parent of " << curNodeID << " does not exist." << endl;
      cerr << "There is an error in containedInGraph/PARENT_NODE_ATTRIBUTE "
	   << "hierarchy." << endl;
      cerr << "findHierarchyPosition is returning erronious values." << endl;
      foundIn = node (NO_PARENT);
      return Coord (0, 0, 0);
    }//end if
	
    //set up the next current node and its contained graph
    curNode = node (curNodeID);
    containedIn = containedInGraph->getNodeValue (curNode);
  }//end while

  //do the last translation at the current level and return the result
  positionOfCur += featureLayout->getNodeValue (curNode);
  foundIn = curNode;
  return positionOfCur;
}//end findHierarchyPosition


//A function to recursively set the spoke order around all branch
//roots.  Thetas will the the theta polar coordinate of the nodes at the
//level below the root.
void TopoLayout::setSpokeOrderAroundRoot (node root) {
  MetricProxy *aveAngle = 
    featureGraph->getLocalProperty<MetricProxy> ("aveAngle");
  aveAngle->setAllNodeValue (0);
  SuperGraph *rootSubGraph = hierarchyMetaGraphs->getNodeValue (root);
  if (rootSubGraph == NULL) return;
  SizesProxy *localBounding = 
    featureGraph->getLocalProperty <SizesProxy> ("boundingBox");
  Size rootSize = localBounding->getNodeValue (root);
  Coord rootCentre (rootSize.getW(), rootSize.getH(), 1.0f);
  rootCentre = rootCentre/2.0;
  
  node maxEdgeNode;
  unsigned int maxEdges = 0;
  vector<node> vecChildren;
  Iterator<node> *children = featureGraph->getOutNodes(root);
  while (children->hasNext()) {
    node curChild = children->next();
    vecChildren.push_back (curChild);
    edge rootEdge = featureGraph->existEdge (root, curChild);
    if (rootEdge.isValid()) {
      const vector<edge> &orgEdge =  originalEdges.get (rootEdge.id);
      if (maxEdges < orgEdge.size()) {
	maxEdges = orgEdge.size();
	maxEdgeNode = curChild;
      }//end if
      double sumAngle = 0;
      vector<edge>::const_iterator orgEdgeIt = orgEdge.begin();
      for (; orgEdgeIt != orgEdge.end(); ++orgEdgeIt) {
	node source = finalLayoutGraph->source (rootEdge);
	node target = finalLayoutGraph->target (rootEdge);
	node sourceMetaNode, targetMetaNode;
	Coord sourcePos = findHierarchyPosition (source, sourceMetaNode);
	Coord targetPos = findHierarchyPosition (target, targetMetaNode);
	Coord rootVec;
	if (sourceMetaNode == root) rootVec = sourcePos - rootCentre;
	else rootVec = targetPos - rootCentre;
	rootVec /= rootVec.norm();
	sumAngle += acos(rootVec.getX())*((rootVec.getY() < 0) ? -1 : 1);
      }//end for
      double rootAveAngle = sumAngle/static_cast<double> (orgEdge.size());
      aveAngle->setNodeValue (curChild, rootAveAngle);
    }//end if
  }//end for
  LessThan aveAngleSort;
  aveAngleSort.metric = aveAngle;
  sort (vecChildren.begin(), vecChildren.end(), aveAngleSort);
  vector<edge> childrenEdges;
  vector<node>::iterator childIt = vecChildren.begin();
  for (; childIt != vecChildren.end(); ++childIt) {
    edge childItEdge = featureGraph->existEdge (root, *childIt);
    childrenEdges.push_back (childItEdge);
  }//end for
  Iterator<edge> *inEdge = featureGraph->getInEdges(root);
  if (featureGraph->indeg (root) == 1)
    childrenEdges.push_back (inEdge->next());
  delete inEdge;
  featureGraph->delLocalProperty ("aveAngle");
  featureGraph->setEdgeOrder (root, childrenEdges);
  children = featureGraph->getOutNodes(root);
  while (children->hasNext()) setSpokeOrderAroundRoot (children->next());
  delete children;
}//end setSpokeOrderAroundRoot

//This function stops the clock timer, reports the time, and
//add the time to the appropriate timer.
inline void stopClock (string type, clock_t start, 
		       double &cumulativeTime) {
  clock_t end = clock();
  double elapsedTime = (double) (end - start) / (double) (CLOCKS_PER_SEC);
  cout << "Time spent on " << type << " " << elapsedTime << endl;
  cumulativeTime += elapsedTime;
}//end stopClock

//This function does the passed layout of the current level
//of the hierarchy
bool TopoLayout::
performLayout (const string &algorithm, clock_t start, 
	       double &cumulativeTime) {
  //perform layout and record the time
  LayoutProxy localLayout(featureGraph);
  DataSet arguments;
  arguments.set<float> ("layer spacing", 1.0);
  arguments.set<float> ("node spacing", 1.0);
  arguments.set<bool> ("bounding circles", true);
  StringCollection tmp ("vertical;");
  arguments.set<StringCollection> ("orientation",tmp); 
  SizesProxy *localBounding = 
    featureGraph->getProperty <SizesProxy> ("boundingBox");
  arguments.set<SizesProxy *> ("boundingBox", localBounding);
  string errMsg;
  bool success = featureGraph->computeProperty (algorithm, &localLayout, 
						errMsg, NULL, &arguments);

  //if the layout was not successful, report the error
  if (!success) {
    cerr << "There was an error while calling the layout function on " 
	 << featureGraph->getAttribute <string> ("name") 
	 << ":\n" << errMsg << endl;
    return false;
  }//end if
  stopClock (featureGraph->getAttribute <string> ("name"), start, 
	     cumulativeTime);

  //set the layout positions
  Iterator <node> *nodes = featureGraph->getNodes ();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    featureLayout->setNodeValue (curNode,
				 localLayout.getNodeValue (curNode));
  } delete nodes; //end while
  Iterator <edge> *edges = featureGraph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();
    featureLayout->setEdgeValue (curEdge,
				 localLayout.getEdgeValue (curEdge));
  } delete edges; //end while
  return true;
}//end performLayout

//A function to layout general graphs using their LayoutAlgorithm object
bool TopoLayout::graphLayout (LayoutAlgorithm &layoutAlgorithm) {
  clock_t graphStart = clock();
  bool success = false;
  bool highPassed = layoutAlgorithm.pluginHigh != "";
  bool lowPassed = layoutAlgorithm.pluginLow != "";
  //if there is a high low test function and both high and low
  //plugins have been passed run the test to determine which plugin
  //to execute.  Otherwise execute the one passed.
  if (layoutAlgorithm.highLowTest && highPassed && lowPassed) {
    if (layoutAlgorithm.highLowTest (featureGraph, this))
      success = performLayout (layoutAlgorithm.pluginHigh,
			       graphStart, layoutAlgorithm.time);
    else 
      success = performLayout (layoutAlgorithm.pluginLow,
			       graphStart, layoutAlgorithm.time);
  }//end if
  else if (highPassed)
    success = performLayout (layoutAlgorithm.pluginHigh,
			     graphStart, layoutAlgorithm.time);
  else if (lowPassed)
    success = performLayout (layoutAlgorithm.pluginLow, 
			     graphStart, layoutAlgorithm.time);
  else {
    cerr << "Error in graphLayout function." << endl;
    cerr << "Neither high or low edge node ratio passed." << endl;
  }//end else
  return success;
}//end graphLayout

//This function selects the appropriate layout algorithm
//and preforms layout of the graph at this level.  It returns true
//if layout succeeds without error.
bool TopoLayout::layout () {
  //for successess/failure of drawing algorithm  
  bool success = false;
  string name = featureGraph->getAttribute<string> ("name");
  string topoType = stripTopoType (featureGraph);
  if (layoutAlgorithmTable.find (topoType.c_str()) != 
      layoutAlgorithmTable.end()) {
    LayoutAlgorithm &layoutAlgo = layoutAlgorithmTable[topoType.c_str()];
    success = graphLayout (layoutAlgo);
    if (!success) {
      cerr << "Error while drawing " << name << endl;
      return false;
    }//end if
  }//end if
  else {
    cerr << "The function to draw " << name 
	 << " was not found in the layout algorithm table." << endl;
    return false;
  }//end else
  return success;
}//end layout

//an inline function to help prevent tortional oscillations
inline double attenu_tau (double tau, double prevTau, double time) {
  //try to do GEM Frick style damping for torques
  if ((tau >= 0) && (prevTau >= 0)) return tau;
  if ((tau <= 0) && (prevTau <= 0)) return tau;
  return tau*time;
}//end attenu_tau

//a member function to compute torque
inline double TopoLayout::computeTorque (node radNode, node opposite, 
					 Coord radBase, double rad) {
  Coord f = prevPos.get (opposite.id) - prevPos.get (radNode.id);
  f.setZ (0);
  if (f.norm() > 1e-6) f /= f.norm();
  Coord r = prevPos.get (radNode.id) - radBase;
  r.setZ (0);
  if (r.norm() > 1e-6) r /= r.norm();
  double cross = r.getX()*f.getY() - r.getY()*f.getX();
  double dot = r.getX()*f.getX() + r.getY()*f.getY();
  return ((prevPos.get (radNode.id) - radBase).norm()/rad)*
    (M_PI/2.0)*((cross < 0) ? -1 : 1)*fabs(dot - 1);
}//end computeTorque

//the following applies torques to the quotient graph nodes
//in order to prevent edge overlap with elements of the quotient
//graph
void TopoLayout::applyTorques () {
  //if there are no edges in this hierarchy there will be no torques
  if (featureGraph->numberOfEdges () == 0) return;

  cerr << "Applying Torques" << endl;

  //get the sizes of the nodes at this level
  SizesProxy *hierarchyNodeSizes = 
    featureGraph->getProperty <SizesProxy> ("viewSize");

  //reset all the the torque values at this level to zero.
  MetricProxy *hierarchyNodeRotations =
    featureGraph->getLocalProperty <MetricProxy> ("viewRotation");
  hierarchyNodeRotations->setAllNodeValue (0);

  //This contains the previous torque on this node
  MutableContainer <double> prevTau;
  prevTau.setAll (0);

  //This contains the initial positions of the nodes involved in
  //torque computations and the node that contains them
  MutableContainer <Coord> initialPositions;
  MutableContainer <node> insideNode;

  //back up the initial positions of the nodes at the ends of torque edges
  Iterator <edge> *initialBackupNodeEdges = featureGraph->getEdges();
  while (initialBackupNodeEdges->hasNext()) {
    edge curEdge = initialBackupNodeEdges->next();
    node curSource = featureGraph->source (curEdge);
    node curTarget = featureGraph->target (curEdge);

    //go through the list of original edges
    vector<edge> backupOrigEdges = originalEdges.get (curEdge.id);
    for (unsigned int j = 0; j < backupOrigEdges.size(); j++) {
      //get the endpoints of the original edge
      node source, target;
      source = finalLayoutGraph->source (backupOrigEdges[j]);
      target = finalLayoutGraph->target (backupOrigEdges[j]);
      node sourceHierarchy, targetHierarchy;
      Coord sourcePosition, targetPosition;
      sourcePosition = findHierarchyPosition (source, sourceHierarchy);
      targetPosition = findHierarchyPosition (target, targetHierarchy);

      //back up the initial positions of the nodes
      initialPositions.set (source.id, sourcePosition);
      initialPositions.set (target.id, targetPosition);
      insideNode.set (source.id, sourceHierarchy);
      insideNode.set (target.id, targetHierarchy);
    }//end for j
  } delete initialBackupNodeEdges;

  //get only the nodes that have a sub hierarchy
  vector <node> hierarchyNodes;
  Iterator <node> *potentialHierarchyNodes = featureGraph->getNodes ();
  while (potentialHierarchyNodes->hasNext()) {
    node curHierarchyNode = potentialHierarchyNodes->next();
    SuperGraph *curHierarchySubGraph = 
      hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
    if (curHierarchySubGraph != NULL) 
      hierarchyNodes.push_back (curHierarchyNode);
  } delete potentialHierarchyNodes; //end while

  //we compute the torque over all nodes of the quotent graph, numNodes times
  //if numNodes below a threshold, push iterations up to the threshold.
  //However, if there are no hierarchy nodes, we don't need to do any
  //torques
  unsigned int iterations = 
    (hierarchyNodes.size() < 5) ? 5 : hierarchyNodes.size();
  if (hierarchyNodes.empty()) iterations = 0;
  for (unsigned int i = 0; i < iterations; i++) {
    //back up the initial positions of the nodes at the ends of torque edges
    StableIterator <edge> backupNodeEdges (featureGraph->getEdges());
    while (backupNodeEdges.hasNext()) {
      edge curEdge = backupNodeEdges.next();
      
      //go through the list of original edges
      vector <edge> backupOrigEdges = originalEdges.get (curEdge.id);
      for (unsigned int j = 0; j < backupOrigEdges.size(); j++) {
	//get the endpoints of the original edge
	node source, target;
	source = finalLayoutGraph->source (backupOrigEdges[j]);
	target = finalLayoutGraph->target (backupOrigEdges[j]);
	
	//do the rotation to the new back up position
	prevPos.set (source.id, rotate (hierarchyNodeRotations
					->getNodeValue 
					(insideNode.get (source.id)),
					initialPositions.get (source.id) -
					featureLayout->getNodeValue 
					(insideNode.get (source.id))) + 
		     featureLayout->getNodeValue 
		     (insideNode.get (source.id)));
	prevPos.set (target.id, rotate (hierarchyNodeRotations
					->getNodeValue 
					(insideNode.get (target.id)),
					initialPositions.get (target.id) -
					featureLayout->getNodeValue 
					(insideNode.get (target.id))) + 
		     featureLayout->getNodeValue 
		     (insideNode.get (target.id)));
      }//end for j
    } //end while

    //go through all the edges of the hierarchy at this level
    for (unsigned int j = 0; j < hierarchyNodes.size(); j++) {
      node curHierarchyNode = hierarchyNodes[j];
      
      //get the translation vector to the position of the quotient
      //graph node for this root graph node.
      Coord hierarchyTrans = featureLayout->getNodeValue(curHierarchyNode);

      //get the radius of the bounding circle of the quotent node
      double rad;
      Size nodeSize = hierarchyNodeSizes->getNodeValue (curHierarchyNode);
      rad = sqrt (sqr(nodeSize.getW()/2.0) + sqr(nodeSize.getH()/2.0));
    
      //go inside the quotient graph and get its inside nodes here somehow.
      //the inside subgraph of the quotient node is quotientSubGraph
      //get those nodes and put them inside insideNodes
      //get a layout proxy and put it inside insideNodesLayout
      SuperGraph *curHierarchySubGraph = 
	hierarchyMetaGraphs->getNodeValue (curHierarchyNode);
     
      //if it is null there is no inside subgraph and no torque is done
      if (curHierarchySubGraph == NULL) continue;
      
      //otherwise, we need to apply the torque all edges involved in
      //a hierarchy edge
      double tau = 0; //the total torque on this node of the hierarchy
      unsigned int numTorqueEdges = 0; //the number of edges contributing
                                       //to the torque
      StableIterator <edge> curHierarchyNodeEdges 
	(featureGraph->getInOutEdges (curHierarchyNode));
      while (curHierarchyNodeEdges.hasNext()) {
	edge curEdge = curHierarchyNodeEdges.next();

	//go through the list of original edges
	vector <edge> curHierarchyNodeOrigEdges =
	  originalEdges.get (curEdge.id);
	for (unsigned int j = 0; j < curHierarchyNodeOrigEdges.size(); j++) { 
	  node source = 
	    finalLayoutGraph->source (curHierarchyNodeOrigEdges[j]);
	  node target =
	    finalLayoutGraph->target (curHierarchyNodeOrigEdges[j]);

	  //determine which is inside the current hierarchy node and apply
	  //the torque.
	  if (curHierarchyNode == insideNode.get (source.id)) 
	    tau += computeTorque (source, target, hierarchyTrans, rad);
	  else 
	    tau += computeTorque (target, source, hierarchyTrans, rad);
	  
	  numTorqueEdges++;
	}//end for j
      } //end while

      //the time to go as %
      double time = (iterations - i) / static_cast <double> (iterations);

      //if there are more than zero edges contributing apply the torque
      //compute the average
      if (numTorqueEdges > 0)
	tau /= static_cast <double> (numTorqueEdges);
      
      //dampen to prevent oscillations
      tau = attenu_tau (tau, prevTau.get(curHierarchyNode.id), time);

      //record the new position of the rotated node
      double cumulativeRotation =
	hierarchyNodeRotations->getNodeValue (curHierarchyNode) + toDeg(tau);
      if (cumulativeRotation > 360) 
	cumulativeRotation -= 360*floor(cumulativeRotation/360);
      if (cumulativeRotation < 0)
	cumulativeRotation += 360*floor(-cumulativeRotation/360 + 1);
      hierarchyNodeRotations->setNodeValue (curHierarchyNode, 
					    cumulativeRotation); 
      //record the applied torque
      prevTau.set (curHierarchyNode.id, tau);

    } //end for j
  }//end for i
  cerr << "done torques " << endl;
}//end applyTorques

//The following function sets the final positions of the nodes in the
//final layout graph.
#define LOWER_EDGE_BY (-5.0)
void TopoLayout::setFinalPositions (SuperGraph *hierarchy,
				    double cumulativeRotation,
				    Coord cumulativeTrans,
				    unsigned int level) { 
  LayoutProxy *finalLayout =
    finalLayoutGraph->getProperty <LayoutProxy> ("viewLayout");
  SizesProxy *finalLayoutSizes =
    finalLayoutGraph->getProperty <SizesProxy> ("viewSize");

  LayoutProxy *hierarchyLayout =
    hierarchy->getLocalProperty<LayoutProxy> ("viewLayout");
  MetricProxy *hierarchyRotation =
    hierarchy->getLocalProperty<MetricProxy> ("viewRotation");
  MetricProxy *hierarchyScale =
    hierarchy->getLocalProperty<MetricProxy> ("viewScale");

  //fix all the bends in the edges at this level
  featureGraph = hierarchy;
  genOriginalEdges();
  Iterator<edge> *edges = hierarchy->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();
    const vector<Coord> bends = hierarchyLayout->getEdgeValue (curEdge);
    const vector<edge> &orgEdge = originalEdges.get (curEdge.id);
    vector<edge>::const_iterator orgEdgeIt = orgEdge.begin();
    for (; orgEdgeIt != orgEdge.end(); ++orgEdgeIt) {
      vector<Coord> orgEdgeBends;
      orgEdgeBends.push_back 
	(Coord (0, 0, (hierarchyDepth - level)*LOWER_EDGE_BY));
      vector<Coord>::const_iterator bendsIt = bends.begin();
      for (; bendsIt != bends.end(); ++bendsIt) {
	Coord rotCoord = rotate (cumulativeRotation, *bendsIt);
	Coord curTrans = cumulativeTrans + rotCoord;
	curTrans.setZ (bendsIt->getZ() + 
		       (hierarchyDepth - level)*LOWER_EDGE_BY);
	orgEdgeBends.push_back (curTrans);
      }//end for
      orgEdgeBends.push_back 
	(Coord (0, 0, (hierarchyDepth - level)*LOWER_EDGE_BY));
      finalLayout->setEdgeValue (*orgEdgeIt, orgEdgeBends);
    }//end for
  } delete edges;
  originalEdges.setAll (vector <edge> ());

  //go through all the nodes at this level
  Iterator<node> *nodes = hierarchy->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();

    Coord pos = hierarchyLayout->getNodeValue (curNode);
    double rotation = hierarchyRotation->getNodeValue (curNode);
    double scale = hierarchyScale->getNodeValue (curNode);

    //transform the node into this coordinate frame
    Coord rotCoord = rotate (cumulativeRotation, pos);
    Coord curTrans = cumulativeTrans + rotCoord;

    //if this node does not have a subgraph associated with it
    //translate it to its final position and scale
    SuperGraph *subHierarchy = 
      hierarchyMetaGraphs->getNodeValue (curNode);
    if (subHierarchy == NULL) {
      finalLayout->setNodeValue (curNode, curTrans);
      Iterator<edge> *nodeEdges = finalLayoutGraph->getInOutEdges (curNode);
      while (nodeEdges->hasNext()) {
	edge curEdge = nodeEdges->next();
	vector<Coord> bends = finalLayout->getEdgeValue (curEdge);
	if (finalLayoutGraph->source(curEdge) == curNode) {
	  bends[0].setX (curTrans.getX());
	  bends[0].setY (curTrans.getY());
	}//end if
	else {
	  bends[bends.size() - 1].setX (curTrans.getX());
	  bends[bends.size() - 1].setY (curTrans.getY());
	}//end else
	finalLayout->setEdgeValue (curEdge, bends);
      } delete nodeEdges;
    }//end if
   
    //Otherwise, get the rotation factor that has been applied to this node
    //and recurse on it.
    else setFinalPositions (subHierarchy, 
			    rotation + cumulativeRotation, curTrans,
			    level + 1);
 
  }delete nodes; //end while
}//end setRealPositions 

//The following function sets the node size of the hierarchy node n
void TopoLayout::setNodeSize (node n, SuperGraph *containedIn) {
  //the sizes of the individual feature nodes
  SizesProxy *featureSizes = 
    featureGraph->getProperty <SizesProxy> ("viewSize");
  MetricProxy *featureRotation =
    featureGraph->getLocalProperty <MetricProxy> ("viewRotation");
  pair<Coord, Coord> boundingBox =
    tlp::computeBoundingBox (featureGraph, featureLayout, featureSizes, 
			     featureRotation);
  Coord boxDims = boundingBox.first - boundingBox.second;
  pair<Coord, Coord> boundingRad =
    tlp::computeBoundingRadius (featureGraph, featureLayout, featureSizes, 
				featureRotation);
  //radius is corner of square.
  double diamLength = M_SQRT2*(boundingRad.second - boundingRad.first).norm();

  //set the size of the node at this level of the hierarchy
  SizesProxy *containedInSizes = 
    containedIn->getLocalProperty <SizesProxy> ("viewSize");
  containedInSizes->setNodeValue (n, Size (diamLength, diamLength, 1.0f));
  SizesProxy *localBounding = 
    containedIn->getLocalProperty <SizesProxy> ("boundingBox");
  localBounding->setNodeValue (n, Size (boxDims.getX(), boxDims.getY(), 1.0f));
}//end setNodeSizes

//function that reports the final times of the components in the algorithm
//start -- the time the plugin began executing
inline void TopoLayout::reportTimes (clock_t start, time_t startLongTime) {
  clock_t end = clock();
  time_t endLongTime;
  time (&endLongTime);
  double diff, diffLong;
  diff = ((double) (end - start) / (double) CLOCKS_PER_SEC);
  diffLong = difftime (endLongTime, startLongTime);
  cout << "-------------------------------------------------------" << endl;
  cout << "Total execution time for layout " << diff << endl;
  cout << "In long time " <<  diffLong << endl; 
  cout << "-------------------------------------------------------" << endl;
  cout << "Trees " << layoutAlgorithmTable[TREE_IDENT].time << endl;
  cout << "Singletons " << layoutAlgorithmTable[SINGLETONS_IDENT].time 
       << endl;
  cout << "Biconnected Components " 
       << layoutAlgorithmTable[BICONNECTED_IDENT].time << endl;
  cout << "Connected Components " 
       << layoutAlgorithmTable[CONNECTED_IDENT].time << endl;
  cout << "Strength Components " 
       << layoutAlgorithmTable[STRENGTH_IDENT].time << endl;
  cout << "Embedder Components "
       << layoutAlgorithmTable[EMBEDDER_IDENT].time << endl;
  cout << "LGL Components " 
       << layoutAlgorithmTable[LGL_IDENT].time << endl;
  cout << "Unknown Components " 
       << layoutAlgorithmTable[UNKNOWN_IDENT].time << endl;
  cout << "Torques " << timeTorques << endl;
  cout << "Centre and Global " << timeCentreGlobalPos << endl;
  cout << "Fast Overlap Resolution " << timeFastOverlaps << endl;
  cout << "-------------------------------------------------------" << endl;
}//end reportTimes

/**
 * This function computes a centre of the graph weighted by node
 * area.  It is required by torques for them to be computed on
 * large layouts.
 */
void TopoLayout::centreByArea (SuperGraph *feature, 
			       LayoutProxy *featureLayout) {
  SizesProxy *sizes = feature->getProperty<SizesProxy> ("viewSize");

  //compute the centre of the graph by area
  Coord max (-DBL_MAX, -DBL_MAX, 0), min (DBL_MAX, DBL_MAX, 0);
  Iterator<node> *nodes = feature->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Coord pos = featureLayout->getNodeValue (curNode);
    Size size = sizes->getNodeValue (curNode);
    if (max.getX() < pos.getX() + size.getW()/2.0)
      max.setX (pos.getX() + size.getW()/2.0);
    if (min.getX() > pos.getX() - size.getW()/2.0)
      min.setX (pos.getX() - size.getW()/2.0);
    if (max.getY() < pos.getY() + size.getH()/2.0)
      max.setY (pos.getY() + size.getH()/2.0);
    if (min.getY() > pos.getY() - size.getH()/2.0)
      min.setY (pos.getY() - size.getH()/2.0);
  } delete nodes;

  //centre the graph by area
  Coord centre = (max + min)/2.0;
  nodes = feature->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    featureLayout->setNodeValue (curNode,
				 featureLayout->getNodeValue (curNode)
				 - centre);
  } delete nodes;
  Iterator<edge> *edges = feature->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();
    vector<Coord> bends = featureLayout->getEdgeValue (curEdge);
    vector<Coord>::iterator bendsIt = bends.begin();
    for (; bendsIt != bends.end(); ++bendsIt) 
      *bendsIt = *bendsIt - centre;
    featureLayout->setEdgeValue (curEdge, bends);
  } delete edges;
}//end centreByArea

//function to layout an interior node of the hierarchy
bool TopoLayout::layoutFeatureNode (SuperGraph *feature) {
  //set up the feature proxys and graph pointers
  featureGraph = feature;
  featureLayout = 
    featureGraph->getLocalProperty <LayoutProxy> ("viewLayout");
  cerr << "Laying out " 
       << featureGraph->getAttribute <string> ("name") << endl;

  //original edges from hierarchy edges
  genOriginalEdges();

  //layout the graph and run options on it
  bool success = layout ();
  if (!success) return false; 
  centreByArea (featureGraph, featureLayout);

  if (torquesRun) {
    clock_t torquesStart = clock();
    applyTorques ();
    stopClock ("torques", torquesStart, timeTorques);
  }//end if
  
  if (fastOverlapRun) {
    if (isType (featureGraph, UNKNOWN_IDENT) ||
	isType (featureGraph, STRENGTH_IDENT) ||
	isType (featureGraph, EMBEDDER_IDENT)) {
      clock_t fastOverlapsStart = clock();
      performLayout (LAYOUT_FAST_OVERLAP_REMOVAL, fastOverlapsStart,
		     timeFastOverlaps);
      centreByArea (featureGraph, featureLayout);
    }//end if
  }//end if
  originalEdges.setAll (vector <edge> ());
  return true;
}//end layoutFeatureNode

//The entry point for the plugin.
bool TopoLayout::run() {
  string errMsg;

  //start the total time clock
  clock_t startTime = clock();
  time_t startLongTime;
  time (&startLongTime);

  //get the plugin parameters
  if (dataSet != NULL) {
    string layoutOperator;
    bool hasDrawOp = 
      dataSet->get (DATASET_DRAW_OPERATOR, layoutOperator);
    if (hasDrawOp) functionTableFromDrawOperator (layoutOperator);
    else functionTableFromUserPrompts();
    setOtherPassOptions();
  }//end if

  //proxies for features and features contained inside
  hierarchyMetaGraphs =
    superGraph->getProperty<MetaGraphProxy> (FEATURE_NODE_NAME);
  containedInGraph = 
    superGraph->getProperty<MetaGraphProxy> (CONTAINED_IN_GRAPH);

  //layout the hierarchy recursively And set its final positions
  bool success = drawHierarchy ();
  if (!success) return false;
  clock_t centreAndGlobalStart = clock();
  setFinalPositions (hierarchy, 0, Coord (0, 0, 0), 0);
  stopClock ("centre and global", centreAndGlobalStart, 
	     timeCentreGlobalPos);

//apply cluster buster if needed to the final layout
  if (clusterbuster) {
    LayoutProxy clusterBusterProxy (finalLayoutGraph);
    finalLayoutGraph->computeProperty ("Leaf Level Cluster Buster", 
				       &clusterBusterProxy, errMsg);
    LayoutProxy *finalLayout =
      finalLayoutGraph->getProperty <LayoutProxy> ("viewLayout");
    Iterator<node> *nodes = finalLayoutGraph->getNodes();
    while (nodes->hasNext()) {
      node curNode = nodes->next();
      finalLayout->setNodeValue (curNode, 
				 clusterBusterProxy.getNodeValue (curNode));
    } delete nodes; //end while
  }//end if
  
  //report the final runtimes
  reportTimes (startTime, startLongTime);
  return true;
}//end run
