//-*-c++-*-
#ifndef __TOPOLAYOUT_H
#define __TOPOLAYOUT_H

/// TopoLayout.h
/** 
 *
 *  An implementation of layout by topological features.  The plugin
 *  takes a hierarchy created by TopoClustering and computes a layout
 *  based on the passed hierarchy of topological features embedded in
 *  the graph.
 *
 *  HISTORY:
 *
 *  The code was first implemented in Nov/Dec 2004 by Daniel Archambault.
 *  It has been subsequently revised into its current form for release in
 *  May 2006.
 *
 *  NOTES:
 *
 *  If the plugin does not find a hierarchy it calls Topological Feature
 *  Clustering automatically to create one.
 *
 *  AUTHOR:
 *
 *  Daniel Archambault, University of British Columbia, CA: 
 *  Email: archam@cs.ubc.ca
 *  Version 0.1: 08 December 2004: Initial version
 *  Version 0.2: 01 May 2006: Revised version.
 *
 *  LICENCE:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by  
 *  the Free Software Foundation; either version 2 of the License, or     
 *  (at your option) any later version.
 *
 */

#include <cstring>
#include <vector>
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <tulip/TulipPlugin.h>
#include <tulip/MutableContainer.h>
#include <HierarchyMap.h>

class TopoLayout:public Layout { 
public:
  //A variable to store the node/edge threshold
  double nodeEdgeThresh;

  //A function to recursively set the spoke order around all branch
  //roots.  Thetas will the the theta polar coordinate of the nodes at the
  //level below the root.
  void setSpokeOrderAroundRoot (node root);

  inline TopoLayout (const PropertyContext &context):Layout(context) {
    clusterbuster = false;
    fastOverlapRun = false;
    torquesRun = true;
    nodeEdgeThresh = 1.0;
    timeTorques = timeCentreGlobalPos = timeFastOverlaps = 
      timeClusterBuster = 0;
    userPrompts();
  }//end constructor

  inline ~TopoLayout() {}

  bool run();
  inline bool check(std::string &erreurMsg) {
    erreurMsg="";
    return true;
  }//end check
  inline void reset() {}

private:
  //time to layout non topology segements of the algorithm
  double timeTorques, timeCentreGlobalPos, 
    timeFastOverlaps, timeClusterBuster;

  //This contains the previous positions of the internal nodes of the
  //quotient graph
  MutableContainer <Coord> prevPos;

  //a proxy to store the position of the hierarchy meta graphs
  MetaGraphProxy *hierarchyMetaGraphs;

  //a proxy that stores the parent graph which contains this node
  MetaGraphProxy *containedInGraph;

  //This contains a pointer to where the final layout of the graph
  //is stored.
  SuperGraph *finalLayoutGraph;

  //This contains a pointer to the hierarchy
  SuperGraph *hierarchy;

  //This contains a pointer to the graph at the current level
  //and a layout proxy to it
  SuperGraph *featureGraph;
  LayoutProxy *featureLayout;

  //This contains a pointer to the untouched graph (only valid at root)
  SuperGraph *untouched;

  //variables to store whether or not certain passes are run
  bool clusterbuster, fastOverlapRun, torquesRun;

  //When the drawing operator is passed using a plugin, 
  //this contains the topological feature types and algorithms
  StringCollection drawingOperator;

  //This contains a pointer to the list of original edges in the graph
  //that correspond to this edge of the graph
  MutableContainer <std::vector <edge> > originalEdges;

  unsigned int hierarchyDepth;

  //----------------------------------------------------------------------
  /**
   * This section defines the algorithm table -- A hash table which
   * specifies the layout algorithms to be used based on topological
   * feature type
   */
  struct eqTopoType {
    bool operator() (const char *s1, const char *s2) const {
      return std::strcmp (s1, s2) == 0;
    }//end operator()
  };
  
  /**
   * Private class which specifies the layout algorithms to be used
   * for a specific topological feature type.  It also records the time
   * to process that type of topology in time and longTime.  pluginHigh is 
   * executed if the result of HighLowTest is true and pluginLow is 
   * executed otherwise.
   */
  class LayoutAlgorithm {
  public:
    //Plugins to run for this algorithm, time to run in two formats,
    //a test to see which format the data is in.
    std::string pluginHigh;
    std::string pluginLow; 
    double time;
    unsigned int longTime;
    bool (*highLowTest) (SuperGraph *featureGraph, TopoLayout *topoLayout);
    
    LayoutAlgorithm () {
      pluginHigh = pluginLow = "";
      time = longTime = 0;
      highLowTest = NULL;
    }//end constructor
  };
  typedef __gnu_cxx::hash_map<const char *, LayoutAlgorithm,
			      __gnu_cxx::hash<const char *>, 
			      eqTopoType> AlgorithmTable;
  AlgorithmTable layoutAlgorithmTable;
  //----------------------------------------------------------------------
  
  /**
   * The function searches the subgraphs of 'superGraph' (the graph
   * on which the plugin was called) and returns a pointer to the
   * graph with name 'graphName'.  If the graph does not exist, NULL
   * is returned.
   */
  SuperGraph *findGraph (std::string graphName);

  /**
   * The function which recursively lays out the feature hierarchy.
   * A pointer to the root of the feature hierarchy is passed in
   * 'curFeatureGraph'.
   */
  bool recurseOnSubHierarchy (SuperGraph *curFeatureGraph,
			      unsigned int &depth);

  /**
   * The function which determines the root of the hierarchy from
   * the 'superGraph' on which the plugin was called.
   */
  bool drawHierarchy ();

  /**
   * This function is called to perform layout of a level in the hierarchy.
   */
  bool layout ();

  /**
   * The function which applies the appropriate layout algorithm to
   * the topological feature.
   */
  bool graphLayout (LayoutAlgorithm &layoutAlgorithm);

  /**
   * Performs the actual layout by calling the appropriate plugin.
   */
  bool performLayout (const std::string &algorithm, clock_t start, 
		      double &cumulativeTime);

  /**
   * Draws the topological feature and applies additional passes
   * to refine the layout
   */
  bool layoutFeatureNode (SuperGraph *feature);

  /**
   * Function which generates the original edge table at this level
   * of the hierarchy.  The original edge table consists of the original
   * edges in the graph represented by all the meta-edges at this
   * level of the hierarchy.
   */
  void genOriginalEdges ();

  /**
   * Returns the coordinates of a node in the original graph 'originalNode"
   * in the context of the current level.  The node 'foundIn' is the 
   * meta-node at this level which contained the passed 'originalNode.'
   */
  Coord findHierarchyPosition (node originalNode, node &foundIn);

  /**
   * The function computes the centre of the graph by area.  That is,
   * it takes node size, not only node position, into account.  The
   * function is required to peform torques properly.
   */
  void centreByArea (SuperGraph *feature, LayoutProxy *featureLayout);

  /**
   * This function applies torques to the current level.
   */
  void applyTorques ();
  
  /**
   * Formula to compute torque on a meta-node.  RadNode is the node in
   * the original graph which is giving rise to the torque on this meta-node.  
   * Opposite is the other node adjacent to this edge.  RadBase is the 
   * centre of the meta-node at this level.  Rad is the length of the
   * radius between radNode and radBase.
   */
  inline double computeTorque (node radNode, node opposite, Coord radBase,
			       double rad);

  /**
   * Sets the node size of a node at this level which contains the 
   * subgraph 'containedIn.'
   */
  void setNodeSize (node n, SuperGraph *containedIn);

  /**
   * function to construct layout parameters dialogue.
   */
  void userPrompts (); 

  /**
   * Sets the highLow test dependent on the topological feature type.
   */
  void setHighLowTest (const std::string &featureType,
		  LayoutAlgorithm &featurAlgo);
  /**
   * fills out the function table from input collected from the
   * user prompt dialogue.
   */
  void functionTableFromUserPrompts ();

  /**
   * fills out the function table from a passed draw operator.
   */
  void functionTableFromDrawOperator (std::string drawingOperator);

  /**
   * determines which additional passes such as torques, fast overlaps
   * per level, etc. are run
   */
  void setOtherPassOptions ();

  /**
   * function to print out the time required to complete each stage
   * of the algorithm
   */
  inline void reportTimes (clock_t start, time_t startLongTime); 

  /**
   * function to set the final layout positions of the graph.
   */
  void setFinalPositions (SuperGraph *hierarchy,
			  double cumulativeRotation,
			  Coord cumulativeTrans, unsigned int level);
};

/**
 * A test that determines if a tree is "bushy" by comparing
 * depth and degree.  The passed tree is "tree"
 */
bool isBushy (SuperGraph *tree, TopoLayout *topoLayout);

/**
 * A test to determine if the graph has a high node to edge ratio.
 */
bool hasHighNodeEdgeRatio (SuperGraph *graph, TopoLayout *topoLayout);

//defines for node/edge ratio thresholds
#define NODE_EDGE_THRESH_DATASET ("Node/Edge")
#define DEFAULT_NODE_EDGE_THRESH ("1.0")

//These define the dataset entries for each topological feature
//----------------------------------------------------------------------
#define DATASET_BUSHY_TREE ("Bushy Tree")
#define DATASET_DEEP_TREE ("Deep Tree")
#define DATASET_CONNECTED ("Connected")
#define DATASET_EMBEDDER ("Near-Mesh")
#define DATASET_STRENGTH_LOW ("Stren Low")
#define DATASET_STRENGTH_HIGH ("Stren High")
#define DATASET_DEFAULT_LOW ("Unkwn Low")
#define DATASET_DEFAULT_HIGH ("Unkwn High")
#define DATASET_LGL_LOW ("LGL Low")
#define DATASET_LGL_HIGH ("LGL High")
//----------------------------------------------------------------------

//These define the set of layout algorithms to be used for a particular
//topological feature
//----------------------------------------------------------------------
#define LAYOUTS_BUSHY_TREE ("Bubble Tree;RINGS Area;Hierarchical Tree (R-T Extended);GEM (Frick) Area;")
#define LAYOUTS_DEEP_TREE ("Hierarchical Tree (R-T Extended);RINGS Area;Bubble Tree;GEM (Frick) Area;")
#define LAYOUTS_CONNECTED ("Connected Componnent Packing;Circular;GEM (Frick) Area;Connected Componnent Packing;")
#define LAYOUTS_EMBEDDER ("Area HDE;High Dimensional Embedder;FM^3;GEM (Frick) Area;")
#define LAYOUTS_STRENGTH_LOW ("GEM (Frick) Area;GRIP;FM^3;Circular;")
#define LAYOUTS_STRENGTH_HIGH ("Circular;GEM (Frick) Area;GRIP;")
#define LAYOUTS_DEFAULT_LOW ("GEM (Frick) Area;GRIP;FM^3;Circular;")
#define LAYOUTS_DEFAULT_HIGH ("Circular;GEM (Frick) Area;GRIP;")
#define LAYOUTS_LGL_HIGH ("LGL Area;GEM (Frick) Area;")
#define LAYOUTS_LGL_LOW ("GEM (Frick) Area;LGL Area;")
#define LAYOUT_FAST_OVERLAP_REMOVAL ("Fast Overlap Removal")
//----------------------------------------------------------------------
#endif
