#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#if (__GNUC__ < 3)
#include <hash_map>
#else
#include <ext/hash_map>
#endif
#include <time.h>
#include <tulip/MutableContainer.h>
#include <tulip/TulipPlugin.h>
#include <tulip/TlpTools.h>
#include "defs.h"
#include "structs.h"
#include "utils.h"
#include "amg_eigen_computation.h"
#include "read_params.h"

using namespace std;

/** \addtogroup layout */
/*@{*/
/// ACE.cpp -- An adaptation of the ACE algorithm.
/**
 * This plugin implements ACE
 * http://www.research.att.com/~yehuda/index_programs.html --
 * for publication and Code
 *
 * Integration in Tulip by D. Archambault
 *
 */
class ACE : public Layout {
public:
  ACE(const PropertyContext &);
  ~ACE();
  bool run();
private:
  void computeEigenvalues(const int ndims, const vector<node> &nodeOrder,
			  double **evecs, double *evals,
			  double *twoNormSqError);
};

ACE::ACE(const PropertyContext &context) : Layout(context) {
}

ACE::~ACE() {}

unsigned int undirNoMultiDeg (SuperGraph *graph, const node &centreNode) {
  unsigned int deg = 0;
  __gnu_cxx::hash_map <unsigned int, bool> adj;
  adj[centreNode.id] = true;
  Iterator<node> *adjNodes = graph->getInOutNodes(centreNode);
  while (adjNodes->hasNext()) {
    node curNode = adjNodes->next();
    if (adj.find (curNode.id) == adj.end()) {
      ++deg;
      adj[curNode.id] = true;
    }//end if
  } delete adjNodes;
  return deg;
}//end undirNoMultiDeg

//A function to print out the eigenvalues of the laplacian matrix
void ACE::computeEigenvalues (const int ndims, const vector<node> &nodeOrder,
			      double **evecs, double *evals, 
			      double *twoNormSqError) {
  //hash map to convert node id to nodeOrder index
  __gnu_cxx::hash_map <unsigned int, unsigned int> idToIndex;
  for (unsigned int i = 0; i < nodeOrder.size(); i++)
    idToIndex[nodeOrder[i].id] = i + 1;
  double AxValues [ndims + 1][nodeOrder.size()];
  //one eigenvalue for each dimension
  for (int i = 1; i <= ndims; i++) {
    evals[i] = 0;
    //Compute the ralay numerator
    for (unsigned int rows = 0; rows < nodeOrder.size(); rows++) {
      AxValues[i][rows] = superGraph->deg (nodeOrder[rows])*evecs[i][rows + 1];
      Iterator<node> *adjNodes = superGraph->getInOutNodes(nodeOrder[rows]);
      while (adjNodes->hasNext()) {
	node curNode = adjNodes->next();
	unsigned int index = idToIndex[curNode.id];
	AxValues[i][rows] -= evecs[i][index];
      } delete adjNodes;
    }//end for rows
    for (unsigned int j = 0; j < nodeOrder.size(); j++)
      evals[i] += AxValues[i][j]*evecs[i][j + 1];
    //compute the mag of this eigenvector 
    double magEvec = 0;
    for (unsigned int j = 1; j <= nodeOrder.size(); j++)
      magEvec += evecs[i][j]*evecs[i][j];
    evals[i] /= magEvec;
  }//end for i

   //print the eigenvalues
  cerr << "eigenvalues: " << endl;
  for (int i = 1; i <= ndims; i++)
    cerr << "lambda_" << i << ":  " << evals[i] << endl;

  //print the squared error (squared two norm) (||Ax - evals[i]*x||^2)
  for (int i = 1; i <= ndims; i++) {
    twoNormSqError[i] = 0;
    for (unsigned int j = 0; j < nodeOrder.size(); j++) {
      double term = AxValues[i][j] - evals[i]*evecs[i][j + 1];
      twoNormSqError[i] += term*term;
    }//end for j
    cerr << "sqr two norm error " << twoNormSqError[i] << endl;
  }//end for i
}//end computeEigenvalues

unsigned long coarsening_time = 0, power_iteration_time = 0, rqi_time = 0;
int total_negative_edges = 0, total_coarse_edges = 0;
double max_neg_percentage = 0;
int IsWeightedGraph = FALSE;

bool ACE::run() {
  //Here we read in the coarsening method and corsening parameters from the
  //file parameters.txt.  All of the file handeling/graph generation commands
  //are ignored.  The only reason we do this is to set layout parameters.
  string parameters_fname ("/layout/params.txt");
  string path = "/imager/project/tulip/softwarePaperVersions/install-toposubmit/lib//tlp/plugins" + parameters_fname;
  GraphType graphType;
  int graphSize1, graphSize2;
  char graphFileName[100];
  char coordFileName[100];
  int useGdraw;
  int ndims;
  int *dims2show;
  int vmax;
  read_parameters (const_cast<char *> (path.c_str()), 
		   &graphType, &graphSize1, 
		   &graphSize2, graphFileName, coordFileName, &useGdraw, 
		   &ndims, &dims2show, &vmax);

  //For a vertex i in a graph of n nodes where i = 0 ... n - 1,
  //bounds stores at i the start index of node i's directly adjacent
  //nodes and at i + 1 the end index of i's directly adjacent nodes
  //inside the adjacency list 'adacency'.  Node order perserves iterator
  //node order and adjIndex hashes tulip node id to adjbounds index.
  int *bounds = 
    (int *)malloc ((superGraph->numberOfNodes() + 1)*sizeof(int));
  int *adjacency = 
    (int *)malloc ((2*superGraph->numberOfEdges() + 1)*sizeof(int));
  vector<node> nodeOrder;
  MutableContainer<int> adjIndex;

  //make the adjacency matrix in the ACE style
  Iterator<node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    adjIndex.set (curNode.id, nodeOrder.size() + 1);
    nodeOrder.push_back (curNode);
  } delete nodes; //end while
  nodes = superGraph->getNodes();
  int edgeCounter = 0;
  int nodeCounter = 0;
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    bounds[nodeCounter++] = edgeCounter; 
    Iterator<node> *adjNodes = superGraph->getInOutNodes(curNode);
    while (adjNodes->hasNext()) {
      node otherNode = adjNodes->next();
      adjacency[edgeCounter++] = adjIndex.get (otherNode.id);
    } delete adjNodes; //end while
  } delete nodes; //end while
  bounds[nodeCounter] = edgeCounter;

  //load the edge weights or initialize them to 1 for unweighted graphs
  //MIGHT WANT TO ALLOW METRIC PROXY INPUT HERE
  float *edgeWeights = (float *) malloc ((edgeCounter + 1)*sizeof(float));
  for (int i = 0; i <= edgeCounter; i++) edgeWeights[i] = 1.0;
  
  //this is the vertex weights used in the mass matrix
  double *nodeWeights = (double *) malloc ((nodeCounter + 1)*sizeof(double));
  for (int i = 0; i <= nodeCounter; i++) nodeWeights[i] = 1.0;

  //compute the eigenvectors
  double **evecs;
  evecs = (double **) malloc(sizeof(double *) * (ndims + 1));
  for (int i=0; i <= ndims; i++) 
    evecs[i]=(double*) malloc((nodeCounter + 1)*sizeof(double));
  struct vtx_data **graph;
  
  time_t startLongTime, endLongTime;
  clock_t startTime = clock();
  time (&startLongTime);

  reformat (bounds, adjacency, nodeCounter, &edgeCounter, NULL, 
	    edgeWeights, &graph);
  if (!useGdraw) {
    delete [] bounds;
    bounds = NULL;
    delete [] adjacency;
    adjacency = NULL;
    if (edgeWeights != NULL)
      delete [] edgeWeights;
    edgeWeights = NULL;
  }
  amg_eigen_computation (graph, nodeCounter, edgeCounter, evecs, ndims, vmax, 
			 0, nodeWeights, levels_between_refinement);

   clock_t endTime = clock();
   time (&endLongTime);
   double diff, diffLongTime;
   diff = (double) (endTime - startTime) / (double) (CLOCKS_PER_SEC);
   diffLongTime = difftime (endLongTime, startLongTime);
  cerr << "time is " << diff << endl;
  cerr << "In long time " << diffLongTime << endl;

  //update the layout using the eigenvectors
  for (int i = 1; i <= nodeCounter; i++) {
    layoutProxy->setNodeValue (nodeOrder[i - 1], 
			       Coord (nodeCounter*evecs[1][i], 
				      nodeCounter*evecs[2][i], 0));
  }
  
  //memory cleanup
  for (int i=1; i<=ndims; i++) free(evecs[i]);
  free (evecs);
  if (nodeWeights != NULL) free (nodeWeights);
  if (edgeWeights != NULL) free (edgeWeights);
  if (graph != NULL) free_graph(graph);
  if (bounds != NULL) free (bounds);
  if (adjacency != NULL) free (adjacency);
  return true;
}

LAYOUTPLUGIN(ACE, "ACE", "Daniel Archambault", "20/06/2005", "Ok", "0", "1");
