#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <tulip/SuperGraph.h>
#include <tulip/TulipPlugin.h>
#include <tulip/Delaunay.h>

#include "EmbedderEmbedGraph.h"
#include "EmbedderParameters.h"
#include "EmbedderDijkstra.h"

#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
   using stdext::hash_map;
#endif

#include <vector>

using namespace std;

/** \addtogroup layout */
/*@{*/
/// Embedder.cpp - An implementation of the Embedder algorithm.
/**
 * This plugin implements ACE:
 * "ACE a fast multiscale eigenvectors computation for drawing huge graphs"
 *
 * Authors: Y. Koren, L. Carmel and D. Harel
 * From the sources of Embedder (reference implementation of the mentionned article)
 *
 * See http://www.wisdom.weizmann.ac.il/~yehuda/ for both publication and source code
 *
 * Integration in Tulip by B.Mathieu
 */

class Embedder : public Layout {
public:
  Embedder(const PropertyContext &);
  ~Embedder();
  bool run();
private:
  vector<node> nodemap;
  hash_map<node, unsigned int> rmap;
};

Embedder::Embedder(const PropertyContext &context) : Layout(context) {
}

Embedder::~Embedder() {}

bool Embedder::run() {
  unsigned int numberOfNodes = (unsigned int) superGraph->numberOfNodes();
  vtx_data *graph = new vtx_data[numberOfNodes];
  int *edges = new int[2 * superGraph->numberOfEdges() + numberOfNodes];
  float *ewgts = new float[2 * superGraph->numberOfEdges() + numberOfNodes];
  double *evals = new double[max_dim];

  nodemap.resize(numberOfNodes);
  rmap.clear();
  rmap.resize(numberOfNodes);

  Iterator<node> *it = superGraph->getNodes();
  unsigned int counter = 0;
  while (it->hasNext()) {
    node n = it->next();
    nodemap[counter] = n;
    rmap[n] = counter++;
  } delete it;

  for (unsigned int i=0; i < numberOfNodes; ++i) {
    node n = nodemap[i];
    graph[i].nedges = superGraph->deg(n) + 1;
    graph[i].edges = edges++;
    graph[i].ewgts = ewgts++;
    graph[i].edges[0] = i;
    graph[i].ewgts[0] = 1.0;
    
    Iterator<node> *tmp = superGraph->getInOutNodes(n);
    while (tmp->hasNext()) {
      node tmpn = tmp->next();
      *edges = rmap[tmpn];
      *ewgts = 1.0;
      ++edges;
      ++ewgts;
    } delete tmp;
  }

  clock_t startTime = clock();
  time_t startLongTime, endLongTime;
  time (&startLongTime);
  DistType **coords = NULL;
  double **dcoords = new double *[max_dim];
  dcoords[0] = new double[numberOfNodes*max_dim];
  for (unsigned int i = 1; i < max_dim; ++i)
    dcoords[i] = &(dcoords[0][numberOfNodes*i]);
  int dims = 50;
  
  embed_graph(graph, numberOfNodes, dims, coords);
  center_coordinate(coords, numberOfNodes, dims);
  PCA(coords, dims, numberOfNodes, dcoords, true, evals);
  clock_t endTime = clock();
  time (&endLongTime);
  double diff, diffLongTime;
  diff = (double) (endTime - startTime) / (double) (CLOCKS_PER_SEC);
  diffLongTime = difftime (endLongTime, startLongTime);
  cerr << "time is " << diff << endl;
  cerr << "long time is " << diffLongTime << endl;
  for (hash_map<node, unsigned int>::const_iterator it = rmap.begin(); it != rmap.end(); ++it) {
    unsigned int index = it->second;
    layoutProxy->setNodeValue(it->first, Coord(dcoords[0][index], 
					       dcoords[1][index]));
  }
  
  if (dataSet != NULL) {
    vector<double> vecEvals;
    vector<vector<double> > vecEvecs (max_dim);
    for (int i = 0; i < max_dim; ++i) {
      vecEvals.push_back (evals[i]);
      for (unsigned int j = 0; j < numberOfNodes; ++j)
	vecEvecs[i].push_back (dcoords[i][j]);
    }//end for i
    dataSet->set ("lambdas", vecEvals);
    dataSet->set ("evecs", vecEvecs);
  }//end if

  delete [] graph[0].edges;
  delete [] graph[0].ewgts;
  delete [] graph;
  delete [] evals;
  delete [] coords[0];
  delete [] coords;
  delete [] dcoords[0];
  delete [] dcoords;
  return true;
}

LAYOUTPLUGIN(Embedder, "High Dimensional Embedder", "Daniel Archambault", "24/08/2005", "Ok", "0", "1");
