#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <algorithm>
#include <iostream>
#include <tulip/TulipPlugin.h>
#include <tulip/MutableContainer.h>
#include <tulip/StlFunctions.h>
#ifdef __cplusplus
extern "C" {
#endif
#include "misf.h"
#include "Graphics.h"
#include "DrawGraph.h"
#include "GRIPPlot.h"
extern Point *component_offset;
#ifdef __cplusplus
}
#endif
using namespace std;
/** \addtogroup layout */
/*@{*/
/// GRIP.cpp -- An adaptation of the GRIP algorithm.
/**
 * This plugin implements GRIP
 * See http://www.cs.arizona.edu/~kobourov/GRIP/ for publication and Code
 *
 * Integration in Tulip by D. Archambault
 *
 * NOTE : although mallocs are not the best way to manage dynamic memory,
 * GRIP is written in pure C and all of its internal memory allocations
 * and reallocations use malloc/realloc.  Therefore, every GRIP structure
 * must be initially allocated with malloc instead of new.
 */
class GRIP : public Layout {
public:
  GRIP(const PropertyContext &);
  ~GRIP();
  bool run();
};


GRIP::GRIP(const PropertyContext &context) : Layout(context) {
}

GRIP::~GRIP() {}

class GRIPIndex {
public:
  int graphIndex;
  int nodeIndex;
};

//operators for equality, inequality and less than for GRIPIndex classes
inline bool operator== (const GRIPIndex &lhs, const GRIPIndex &rhs) {
  return (lhs.graphIndex == rhs.graphIndex) && 
    (lhs.nodeIndex == rhs.nodeIndex);
}//end operator==
inline bool operator!= (const GRIPIndex &lhs, const GRIPIndex &rhs) {
    return !(lhs == rhs);
}//end operator!=
inline bool operator< (const GRIPIndex &lhs, const GRIPIndex &rhs) {
  return (lhs.graphIndex < rhs.graphIndex) ||
	  ((lhs.graphIndex == rhs.graphIndex) && 
	   (lhs.nodeIndex < rhs.nodeIndex));
}//end operator<

//This mutable container maps a node ID to a GRIP index, wich consists
//of a graph number and an adjacency row number
MutableContainer<GRIPIndex> nodeIDtoGripIndex;
 
//a comparator to sort nodes by GRIP index
bool compareGRIPIndex (const node n1, const node n2) {
  GRIPIndex n1gi = nodeIDtoGripIndex.get(n1.id);
  GRIPIndex n2gi = nodeIDtoGripIndex.get(n2.id);
  return n1gi < n2gi;
}//end compareGRIPIndex 

bool GRIP::run() {

  //stores a new order on the nodes that is basically the order in the
  //passed GRIP graph data structure
  std::vector<node> nodeOrder;
  typedef std::vector<node>::const_iterator constNodeOrderIt;

  //the graph must have at least one node for the GRIP code to work
  if (superGraph->numberOfNodes() < 1) return true;
  
  //create adjacency structure used by GRIP 
  size_tt **fullAdjList = 
    (size_tt **) malloc((superGraph->numberOfNodes() + 1) * sizeof(size_tt *));
  fullAdjList[0] =
    (size_tt *)malloc(superGraph->numberOfNodes() * sizeof(size_tt));
  Iterator<node> *nodes = superGraph->getNodes();
  unsigned int nodeCounter = 0;
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    nodeOrder.push_back (curNode);
    GRIPIndex curIndex;
    curIndex.graphIndex = 0;
    curIndex.nodeIndex = nodeCounter;
    fullAdjList[0][nodeCounter] = superGraph->deg(curNode);
    fullAdjList[nodeCounter + 1] = 
      (size_tt *)malloc(superGraph->deg(curNode) * sizeof(size_tt));
    nodeIDtoGripIndex.set(curNode.id, curIndex);
    ++nodeCounter;
  } delete nodes; //end while

  //fill the adjacency structure used by grip with the adj matrix
  nodeCounter = 0;
  for (constNodeOrderIt nodeIt = nodeOrder.begin(); nodeIt != nodeOrder.end();
       nodeIt++) {
    node curNode = *nodeIt;
    Iterator<node> *adjNodes = superGraph->getInOutNodes(curNode);
    unsigned int edgeCounter = 0;
    while (adjNodes->hasNext()) {
      node adjNode = adjNodes->next();
      fullAdjList[nodeCounter + 1][edgeCounter++] = 
	(nodeIDtoGripIndex.get(adjNode.id)).nodeIndex;
    } delete adjNodes; //end while
    ++nodeCounter;
  }//end for nodeIt

  time_t startLongTime, endLongTime;
  clock_t startTime = clock();
  time (&startLongTime);

  //compute the number of connected components using the GRIP code
  Graph *graph = (Graph *)malloc(sizeof(Graph));
  graph->numOfVert = superGraph->numberOfNodes();
  graph->adjList = fullAdjList;
  VisualGraphics *grip_graph = 
    (VisualGraphics *) malloc (sizeof (VisualGraphics));
  grip_graph->node_graphics = NULL;
  grip_graph->edge_graphics = NULL;
  grip_graph->graph = graph;
  int **new_position;
  grip_graph = get_graph_components(grip_graph, true, &new_position);

  //if there is more than one component the node indexes in the grip
  //structure have changed.  Update the indexes and reorder nodeOrder.
  if (grip_graph->numComponents > 1) {
    for (unsigned int i = 0; i < graph->numOfVert; i++) {
      GRIPIndex curIndex;
      curIndex.graphIndex = new_position[i][0];
      curIndex.nodeIndex = new_position[i][1];
      nodeIDtoGripIndex.set (nodeOrder[i].id, curIndex);
    }//end for i
    std::sort (nodeOrder.begin(), nodeOrder.end(), compareGRIPIndex);
    for (unsigned int i = 0; i < graph->numOfVert; i++)
      free(new_position[i]);
    free(new_position); 
  }//end if
  grip_graph->edge_graphics = NULL;
  grip_graph->node_graphics = NULL;
  std::cout << "grip comps " << grip_graph->numComponents << std::endl;

  //compute the grip layout on the structures
  int _dim = 2;
  int _numOfInitVert = 4;
  int _displayPar = 1;
  int _random = 0;
  int _FR_full = 0;
  int _FR_levels = 1;
  int _plot_all_vert = 0;    //indicates whether only FR all is to be used
  int _rounds = 20;
  int _finalRounds = 30;
  float _heat_fraction = 0.17;
  float _r = 0.15;  //parameters of update_Local_Temp_v2()
  float _s = 3.0;
  DrawGraph *dg = init_DG(grip_graph->graph, grip_graph->numComponents, _dim,  _numOfInitVert, _displayPar, _random, _FR_full, _FR_levels, _plot_all_vert, _rounds, _finalRounds, _heat_fraction, _r, _s);
  dg->edge = 32;
  for (int i = 0; i < grip_graph->numComponents; ++i) 
    misf_engine(&grip_graph->graph[i], &dg[i]);
  plotGraphNoDraw (grip_graph, dg, 700, 700); 
  for (int i = 0; i < grip_graph->numComponents; ++i) {
    while (dg[i].currLevel > -1) 
      misf_engine(&grip_graph->graph[i], &dg[i]);
  }

  clock_t endTime = clock();
  time (&endLongTime);
  double diff, diffLongTime;
  diff = (double) (endTime - startTime) / (double) (CLOCKS_PER_SEC);
  diffLongTime = difftime (endLongTime, startLongTime);
  cerr << "time is " << diff << endl;
  cerr << "diff long time " << diffLongTime << endl;

  //load the final positions into layoutProxy
  for (constNodeOrderIt nodeIt = nodeOrder.begin(); nodeIt != nodeOrder.end();
       nodeIt++) {
    node curNode = *nodeIt;
    GRIPIndex index = nodeIDtoGripIndex.get(curNode.id);
    layoutProxy->setNodeValue (curNode,
			       Coord (component_offset[index.graphIndex]
				      .coord.d[0] + 
				      dg[index.graphIndex].pos[index.nodeIndex]
				      .coord.d[0],
				      component_offset[index.graphIndex]
				      .coord.d[1] + 
				      dg[index.graphIndex].pos[index.nodeIndex]
				      .coord.d[1], 0.0));
  }//end for comp


  //clean up memory
  free_DrawGraph (dg);
  free_graphics (grip_graph);
  free_memory();
  return true;
}

LAYOUTPLUGIN(GRIP, "GRIP", "Daniel Archambault", "20/06/2005", "Ok", "0", "1");
